package spotai;

import java.util.concurrent.*;
import nintaco.api.*;
import static spotai.BonusItems.*;

public class SpotAI {
  
  private static enum State {
    WaitingForTurn,
    
    Thinking,
    MovingCursorToSource,
    DelayAfterMovingCursorToSource,
    PressingSourceButton,
    ReleasingSourceButton,
    MovingCursorToDestination,
    DelayAfterMovingCursorToDestination,
    PressingDestinationButton,
    DelayAfterPressingDestinationButton,
    
    BonusThinking,
    BonusReelsDelay,
    WatchingBonusReelSpin,
    PressingReelButton,
    ReleasingReelButton,
    WaitingForEndOfBonus,
  }
  
  private static final int[][] BONUS_TARGETS = {
    { 0x0F, 0x05, 0x0C, }, // Spot
    { 0x00, 0x09, 0x10, }, // Cherries
    { 0x06, 0x0D, 0x02, }, // Arcadias
  };
    
  private static final int ROWS = 7;
  private static final int COLUMNS = 7;
  private static final int SHORT_DELAY = 20;
  private static final int LONG_DELAY = 60;
  
  private final Hypermax hypermax = new Hypermax();
  private final API api = ApiSource.getAPI();  
  private final ExecutorService executor = Executors.newFixedThreadPool(1);
  private final boolean[][] topology = new boolean[ROWS][COLUMNS];
  private final Node node = new Node();
  private final boolean[] humanPlayers = new boolean[4];
  private final boolean[] participants = new boolean[4];    
  
  private boolean lastHumanPlaying = true;
  private boolean lastBonusSlots = true;
  private int playerCount;
  private int gamepad;
  private int delay;
  private State state = State.WaitingForTurn;
  
  private volatile boolean moveReady;

  public void launch() {
    node.nodeType = NodeType.Intermediate;
    node.playfield = new int[ROWS][COLUMNS];    
    api.addFrameListener(this::renderFinished);
    api.run();
  }
  
  private void renderFinished() {
    switch(state) {
      case WaitingForTurn: {
        final boolean bonusSlots = readBonusSlots();
        final boolean bonus = !lastBonusSlots && bonusSlots;
        final boolean humanPlaying = readHumanPlaying();
        if (bonus || (!lastHumanPlaying && humanPlaying)) {    
          readPlayfield(); 
          moveReady = false;
          state = bonus ? State.BonusThinking : State.Thinking;
          executor.submit(() -> computeMove(bonus));
        }   
        lastBonusSlots = bonusSlots;
        lastHumanPlaying = humanPlaying;
        break;
      }
        
      case Thinking:
        drawCenteredString("AI Thinking...", 28);
        if (moveReady) {
          state = State.MovingCursorToSource;
        }
        break;
      case MovingCursorToSource:
        moveCursor(node.source);
        break;  
      case DelayAfterMovingCursorToSource:
        if (--delay < 0) {
          delay = SHORT_DELAY;
          state = State.PressingSourceButton;
        }
        break;
      case PressingSourceButton:
        api.writeGamepad(gamepad, GamepadButtons.A, true);
        if (--delay < 0) {
          delay = SHORT_DELAY;
          state = State.ReleasingSourceButton;
        }
        break;
      case ReleasingSourceButton:
        if (--delay < 0) {
          state = State.MovingCursorToDestination;
        }
        break;
      case MovingCursorToDestination:
        moveCursor(node.destination);
        break;
      case DelayAfterMovingCursorToDestination:
        if (--delay < 0) {
          delay = SHORT_DELAY;
          state = State.PressingDestinationButton;
        }
        break;
      case PressingDestinationButton:
        api.writeGamepad(gamepad, GamepadButtons.A, true);
        if (--delay < 0) {
          delay = LONG_DELAY;
          state = State.DelayAfterPressingDestinationButton;
        }
        break;
      case DelayAfterPressingDestinationButton:
        if (--delay < 0) {
          lastBonusSlots = lastHumanPlaying = true;
          node.bonus = None;
          state = State.WaitingForTurn;          
        }
        break;
        
      case BonusThinking:
        drawCenteredString("AI Thinking...", 28);
        if (moveReady) {
          if (node.source == Coordinate.PASS) {
            lastBonusSlots = lastHumanPlaying = true;
            node.bonus = None;
            state = State.WaitingForTurn;
          } else {
            delay = LONG_DELAY;
            state = State.BonusReelsDelay;
          }
        }
        break;
      case BonusReelsDelay:
        if (--delay < 0) {          
          state = State.WatchingBonusReelSpin;
        }
        break;
      case WatchingBonusReelSpin: {
        final int reel = api.readCPU(Addresses.StoppedBonusReels);
        if (api.readCPU(Addresses.BonusReels + reel) 
            == BONUS_TARGETS[node.bonus][reel]) {
          state = State.PressingReelButton;
          delay = SHORT_DELAY;
        }
        break;
      }
      case PressingReelButton:
        api.writeGamepad(gamepad, GamepadButtons.A, true);
        if (--delay < 0) {
          delay = LONG_DELAY;
          state = State.ReleasingReelButton;
        }
        break;
      case ReleasingReelButton:
        if (--delay < 0) {
          if (api.readCPU(Addresses.StoppedBonusReels) < 3) {
            state = State.WatchingBonusReelSpin;
          } else {
            state = State.WaitingForEndOfBonus;
          }
        }
        break;
      case WaitingForEndOfBonus:
        if (!readBonusSlots()) {
          if (node.bonus == Arcadias || node.source == null) {
            state = State.MovingCursorToDestination;
          } else if (readHumanPlaying()) {
            state = State.MovingCursorToSource;
          }
        }
        break;
    }
  }
  
  private void moveCursor(final Coordinate c) {
    final int cursorCell = api.readCPU(Addresses.CursorCell) - 1;
    if (cursorCell < 0 || cursorCell > ROWS * COLUMNS - 1) {
      state = State.WaitingForTurn;
      lastBonusSlots = lastHumanPlaying = true;
      node.bonus = None;
    }
    final int cursorX = cursorCell % ROWS;
    final int cursorY = cursorCell / ROWS;
    if (cursorX > c.x) {
      api.writeGamepad(gamepad, GamepadButtons.Left, true);
    } else if (cursorX < c.x) {
      api.writeGamepad(gamepad, GamepadButtons.Right, true);
    }
    if (cursorY > c.y) {
      api.writeGamepad(gamepad, GamepadButtons.Up, true);
    } else if (cursorY < c.y) {
      api.writeGamepad(gamepad, GamepadButtons.Down, true);
    }
    if (cursorX == c.x && cursorY == c.y) {
      delay = SHORT_DELAY;
      switch(state) {
        case MovingCursorToSource:
          state = State.DelayAfterMovingCursorToSource;
          break;
        case MovingCursorToDestination:
          state = State.DelayAfterMovingCursorToDestination;
          break;
      }      
    }
  }
  
  private void computeMove(final boolean bonus) {
    node.set(hypermax.hypermax(node, 0, bonus));
    if (node.source != Coordinate.PASS) {
      moveReady = true;
    }
  }
  
  private void readPlayfield() {
    
    boolean reset = false;
    int players = 0;
    for(int i = 3; i >= 0; i--) {
      final int type = api.readCPU(Addresses.PlayerTypes + i);

      final boolean participant = i < 2 || type != 0;
      if (participant) {
        players++;
      }
      if (participant != participants[i]) {
        participants[i] = participant;
        reset = true;
      }
      
      final boolean humanPlayer = participant && type != 5;
      if (humanPlayer != humanPlayers[i]) {
        humanPlayers[i] = humanPlayer;
        reset = true;
      }
    }
    
    if (players != playerCount) {
      playerCount = players;
      reset = true;
    }
        
    final int[][] playfield = node.playfield;
    for(int y = ROWS - 1; y >= 0; y--) {
      final int rowAddress = Addresses.Playfield + COLUMNS * y;
      for(int x = COLUMNS - 1; x >= 0; x--) {
        int cell = api.readCPU(rowAddress + x);
        final boolean reachable = (cell & 0xF0) != 0x80;
        if (topology[y][x] != reachable) {          
          topology[y][x] = reachable;
          reset = true;
        }
        cell = (cell & 0x0F) - 1;
        playfield[y][x] = (!participants[2] && cell == 3) ? 2 : cell;
      }
    }
    
    gamepad = api.readCPU(Addresses.CurrentPlayer);
    node.player = (!participants[2] && gamepad == 3) ? 2 : gamepad;
    
    if (reset) {
      node.utilities = new float[playerCount];
      node.bonus = None;
      hypermax.reset(topology, playerCount, playerCount + 1);
    }
    hypermax.evaluate(node);
  }
  
  private boolean readBonusSlots() {
    return api.readCPU(Addresses.Mode) == 0x00 
        && api.readCPU(Addresses.BonusMode) == 0x0F;
  }
  
  private boolean readHumanPlaying() {
    return api.readCPU(Addresses.HumanPlaying) == 0x02;
  }
  
  private void drawCenteredString(final String str, final int y) {
    final int width = api.getStringWidth(str, false) + 1;
    final int x = (255 - width) >> 1;
    api.setColor(Colors.DARK_BLUE);
    api.fillRect(x, y, width, 10);
    api.setColor(Colors.BLUE);
    api.drawRect(x, y, width, 10);
    api.setColor(Colors.WHITE);
    api.drawString(str, x + 1, y + 2, false);
  }
  
  public static void main(final String... args) throws Throwable {
    System.out.println("AI started.");
    ApiSource.initRemoteAPI("localhost", 9999);
    new SpotAI().launch();
  }  
}