package spotai;

import java.io.*;
import java.util.*;
import static java.lang.Math.*;
import static spotai.Hypermax.*;

public class TestHypermax {
  
  private static final int ROWS = 7;
  private static final int COLUMNS = 7;
  
  private final BufferedReader in = new BufferedReader(
      new InputStreamReader(System.in));

  public void launch() throws Throwable {
    final boolean[][] topology = new boolean[ROWS][COLUMNS];
    for(int i = ROWS - 1; i >= 0; i--) {
      Arrays.fill(topology[i], true);
    }
    
    final Hypermax hypermax = new Hypermax();
    while(true) {
      final int playerCount = promptInt("How many players (2--4)? ", 2, 4);
      final Set<Integer> humans = promptHumans(playerCount);
      final int maxDepth = promptInt("Specify difficulty (1--9)? ", 1, 9);            
      hypermax.reset(topology, playerCount, maxDepth);
      
      Node node = new Node();
      node.nodeType = NodeType.Intermediate;
      node.playfield = new int[ROWS][COLUMNS];
      node.utilities = new float[playerCount];
      for(int i = ROWS - 1; i >= 0; i--) {
        Arrays.fill(node.playfield[i], -1);
      }
      switch(playerCount) {
        case 2:
          node.playfield[0][0] = node.playfield[ROWS - 1][COLUMNS - 1] = 0;
          node.playfield[0][COLUMNS - 1] = node.playfield[ROWS - 1][0] = 1;
          break;
        case 3:
          node.playfield[0][0] = 0;
          node.playfield[0][COLUMNS - 1] = 1;
          node.playfield[ROWS - 1][0] = 2;
          break;
        case 4:
          node.playfield[0][0] = 0;
          node.playfield[0][COLUMNS - 1] = 1;
          node.playfield[ROWS - 1][0] = 2;
          node.playfield[ROWS - 1][COLUMNS - 1] = 3;                    
          break;
      }
      
      int player = 0;
      while(true) {
        node.printPlayfield();
        if (humans.contains(player)) {
          promptMove(hypermax, node, player, playerCount);          
        } else {
          System.out.format("Player %d (AI) is thinking...%n", player);
          node.set(hypermax.hypermax(node, 0, false));
          System.out.println();
          if (node.source == Coordinate.PASS) {
            System.out.format("Player %d (AI) was forced to pass.%n", player);
          } else {          
            System.out.format("Player %d (AI) moved from (%d, %d) to (%d, %d)"
                + ".%n", player, node.source.x, node.source.y, 
                    node.destination.x, node.destination.y);
          }           
        }
        if (node.nodeType != NodeType.Intermediate) {
          node.printPlayfield();
          if (node.nodeType == NodeType.Draw) {
            System.out.println("The game ended in a draw.");
          } else {
            System.out.format("Player %d (%s) wins.%n", player, 
                humans.contains(player) ? "Human" : "AI");
          }
          System.out.println();
          break;
        }
        if (++player == playerCount) {
          player = 0;
        }
      }
    }
  }
  
  private void promptMove(final Hypermax hypermax, final Node node, 
      final int player, final int playerCount) {
    
    outer: while(true) {     
      System.out.format("Enter player %d (human) move (x1 y1 x2 y2): ", 
          player);
      final String[] tokens = promptString().split("[^\\d]");
      final List<Integer> list = new ArrayList<>();      
      for(final String token : tokens) {
        if (!token.trim().isEmpty()) {
          int value = -1;
          try {
            value = Integer.parseInt(token);
          } catch(final Throwable t) {          
          }          
          list.add(value);
        }
      }
      if (list.size() != 4) {
        continue outer;
      }
      final int x1 = list.get(0);
      final int y1 = list.get(1);
      final int x2 = list.get(2);
      final int y2 = list.get(3);
      if (x1 < 0 || x1 > COLUMNS || y1 < 0 || y1 > ROWS
          || x2 < 0 || x2 > COLUMNS || y2 < 0 || y2 > ROWS
          || abs(x2 - x1) > 2 || abs(y2 - y1) > 2
          || node.playfield[y1][x1] != player 
          || node.playfield[y2][x2] >= 0) {
        continue outer;
      }
      final int moveType = (abs(y2 - y1) == 2 || abs(x2 - x1) == 2) 
          ? SKIPS : CLONES;
      final Coordinate[] moves = hypermax.getMoves()[y1][x1][moveType];
      inner: {
        for(int i = moves.length - 1; i >= 0; i--) {
          final Coordinate m = moves[i];
          if (m.x == x2 && m.y == y2) {
            break inner;
          }
        }
        continue outer;
      }
      if (moveType == SKIPS) {
        node.playfield[y1][x1] = -1;        
      }
      node.playfield[y2][x2] = player;
      final Coordinate[] cs = hypermax.getMoves()[y2][x2][CLONES];
      for(int i = cs.length - 1; i >= 0; i--){
        final Coordinate c = cs[i];
        final int p = node.playfield[c.y][c.x];
        if (p >= 0 && p != player) {
          node.playfield[c.y][c.x] = player;
        }
      }  
      if (++node.player == playerCount) {
        node.player = 0;
      }
      hypermax.evaluate(node);
      break;
    }
  }
  
  private Set<Integer> promptHumans(final int playerCount) {
    outer: while(true) {
      System.out.format("List the human players (0--%d): ", playerCount - 1);
      final String[] tokens = promptString().split("[^\\d]");
      final Set<Integer> result = new HashSet<>();      
      for(final String token : tokens) {
        if (!token.trim().isEmpty()) {
          int value = -1;
          try {
            value = Integer.parseInt(token);
          } catch(final Throwable t) {          
          }
          if (value < 0 || value >= playerCount) {
            continue outer;
          }
          result.add(value);
        }
      }
      if (result.size() <= playerCount) {
        return result;
      }
    }
  }
  
  private int promptInt(final String message, final int min, final int max) {
    int value = -1;
    do {
      System.out.print(message);      
      value = promptInt();
    } while(value < min || value > max);
    return value;
  }
  
  private int promptInt() {
    try {
      return Integer.parseInt(in.readLine());
    } catch(final Throwable t) {      
    }
    return -1;
  }
  
  private String promptString() {
    try {
      return in.readLine();
    } catch(final Throwable t) {      
    }
    return "";
  }
  
  public static void main(final String... args) throws Throwable {
    new TestHypermax().launch();
  }  
}